#ifndef SPECIAL_KEYS_EVDEV_H_
#define SPECIAL_KEYS_EVDEV_H_

enum {
	RENDERKEY_DOWN = 128,
	RENDERKEY_UP,
	RENDERKEY_LEFT,
	RENDERKEY_RIGHT,
	RENDERKEY_ESC,
	RENDERKEY_DEL,  
	RENDERKEY_LSHIFT, 
	RENDERKEY_RSHIFT,
	RENDERKEY_LCTRL,
	RENDERKEY_RCTRL,
	RENDERKEY_LALT,
	RENDERKEY_RALT,
	RENDERKEY_HOME,
	RENDERKEY_END,
	RENDERKEY_PAGEUP,
	RENDERKEY_PAGEDOWN,
	RENDERKEY_F1,
	RENDERKEY_F2,
	RENDERKEY_F3,
	RENDERKEY_F4,
	RENDERKEY_F5,
	RENDERKEY_F6,
	RENDERKEY_F7,
	RENDERKEY_F8,
	RENDERKEY_F9,
	RENDERKEY_F10,
	RENDERKEY_F11,
	RENDERKEY_F12,
	RENDERKEY_SELECT,
	RENDERKEY_PLAY,
	RENDERKEY_SEARCH,
	RENDERKEY_BACK,
	RENDERKEY_NP0,
	RENDERKEY_NP1,
	RENDERKEY_NP2,
	RENDERKEY_NP3,
	RENDERKEY_NP4,
	RENDERKEY_NP5,
	RENDERKEY_NP6,
	RENDERKEY_NP7,
	RENDERKEY_NP8,
	RENDERKEY_NP9,
	RENDERKEY_NP_DIV,
	RENDERKEY_NP_MUL,
	RENDERKEY_NP_MIN,
	RENDERKEY_NP_PLUS,
	RENDERKEY_NP_ENTER,
	RENDERKEY_NP_STOP,
	RENDER_DPAD_UP,
	RENDER_DPAD_DOWN,
	RENDER_DPAD_LEFT,
	RENDER_DPAD_RIGHT
};

#endif //SPECIAL_KEYS_EVDEV_H_
