	E(CMD_OK),
	E(CMD_VERSION),
	E(CMD_ECHO),
	E(CMD_AP_SCAN),
	E(CMD_AP_CFG),
	E(CMD_AP_CFG_GET),
	E(CMD_IP_CURRENT),
	E(CMD_RESERVED_7),
	E(CMD_IP_CFG),
	E(CMD_IP_CFG_GET),
	E(CMD_DEF_AP_CFG),
	E(CMD_DEF_AP_CFG_GET),
	E(CMD_AP_JOIN),
	E(CMD_AP_LEAVE),
	E(CMD_TCP_CON),
	E(CMD_TCP_BIND),
	E(CMD_RESERVED_16),
	E(CMD_CLOSE),
	E(CMD_UDP_SET),
	E(CMD_RESERVED_19),
	E(CMD_SOCK_STAT),
	E(CMD_PING),
	E(CMD_SNTP_CFG),
	E(CMD_SNTP_CFG_GET),
	E(CMD_DATETIME),
	E(CMD_DT_SET),
	E(CMD_FLASH_WRITE),
	E(CMD_FLASH_READ),
	E(CMD_FLASH_ERASE),
	E(CMD_FLASH_ID),
	E(CMD_SYS_STAT),
	E(CMD_DEF_CFG_SET),
	E(CMD_HRNG_GET),
	E(CMD_BSSID_GET),
	E(CMD_GAMERTAG_SET),
	E(CMD_GAMERTAG_GET),
	E(CMD_LOG),
	E(CMD_FACTORY_RESET),
	E(CMD_SLEEP),
	E(CMD_HTTP_URL_SET),
	E(CMD_HTTP_METHOD_SET),
	E(CMD_HTTP_CERT_QUERY),
	E(CMD_HTTP_CERT_SET),
	E(CMD_HTTP_HDR_ADD),
	E(CMD_HTTP_HDR_DEL),
	E(CMD_HTTP_OPEN),
	E(CMD_HTTP_FINISH),
	E(CMD_HTTP_CLEANUP),
	E(CMD_RESERVED_48),
	E(CMD_SERVER_URL_GET),
	E(CMD_SERVER_URL_SET),
	E(CMD_WIFI_ADV_GET),
	E(CMD_WIFI_ADV_SET),
	E(CMD_NV_CFG_SAVE),
	E(CMD_UPGRADE_LIST),
	E(CMD_UPGRADE_PERFORM),
